// -*- mode: c++; indent-tabs-mode: nil -*-
//! @file SoapClient.qm SOAP Client module implementation based on the WSDL classes

/*  SoapClient.qm Copyright (C) 2012 - 2014 David Nichols

    Permission is hereby granted, free of charge, to any person obtaining a
    copy of this software and associated documentation files (the "Software"),
    to deal in the Software without restriction, including without limitation
    the rights to use, copy, modify, merge, publish, distribute, sublicense,
    and/or sell copies of the Software, and to permit persons to whom the
    Software is furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in
    all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.
*/

// make sure we have the required qore version

// requires XML functionality

// requires the WSDL module

// need mime definitions


/*  Version History
    * 2014-04-22 v0.2.4: David Nichols <david@qore.org>
    *  + handle SOAP fault messages returned with 500-series error codes as Fault messages

    * 2014-03-14 v0.2.4: David Nichols <david@qore.org>
    *  + added optional logging closures, added SOAP headers to info hash when available

    * 2014-03-04 v0.2.4: David Nichols <david@qore.org>
    *  + fixed URI request path, fixed agent string

    * 2012-05-31 v0.2.3: David Nichols <david@qore.org>
    *  + updated to a user module
*/

/** @mainpage SoapClient Module
    
    @section soapclientintro Introduction to the SoapClient Module

    The SoapClient module implements a minimal SOAP client using WSDL, XSD, SOAP support implemented in the WSDL module.

    Create outgoing request messages by setting up a Qore data structure corresponding to the SOAP message.  Exceptions will be thrown if either the outgoing or the response message do not corespond to the WSDL.   The exceptions should be fairly verbose to allow you to quickly correct any mistakes.

    Currently the WSDL implementation is fairly basic so any messages using unimplemented features of SOAP or XSD will fail.

    @section Example

    Here is an example of how to use this module:
    @code
%requires SoapClient
my SoapClient $sc = new SoapClient(("wsdl" : "http://soap.server.org:8080/my-service?wsdl"));
my hash $msg = getMessage();
my any $result = $sc.call("SubmitDocument", $msg);
    @endcode

    The SoapClient::constructor() takes named arguments in the form of a hash; valid arguments are:
    - required keys: one of:
      - \c "wsdl" or \c "wsdl_file": a string defining the WSDL or the URL of the WSDL
    - optional keys:
      - \c "service": the name of the "portType" to use (if more than 1 portType is defined in the WSDL then this key is mandatory
      - \c "url": to override the URL defined in the WSDL
      - \c "headers": to override any HTTP headers sent in outgoing messages
      - \c "event_queue": to set an I/O event queue on the HTTPClient
      - \c "log": a @ref closure "closure" or @ref call_reference "call reference" taking a single string argument for informational logging
      - \c "dbglog": a @ref closure "closure" or @ref call_reference "call reference" taking a single string argument for verbose technical debugging logging
    
    Also the following keys can be set to set HTTP options:
    - \c "connect_timeout"
    - \c "http_version"
    - \c "max_redirects"
    - \c "proxy"
    - \c "timeout"
*/

//! main SoapClient namespace
namespace SoapClient {
    //! SOAP client class implementation, publically inherits qore's HTTPClient class
class SoapClient : public HTTPClient {

public:
        //! version of the implementation of this class
        const Version = "0.2.4";

        //! default HTTP headers
        const Headers = ("Accept": (MimeTypeSoapXml + "," + MimeTypeXml + "," + MimeTypeXmlApp), "User-Agent": ("Qore-Soap-Client/" + SoapClient::Version));

        //! option keys passed to the HTTPClient constructor
        const HTTPOptions = ( "connect_timeout", "http_version", "max_redirects", "proxy", "timeout" );

        //! @cond nodoc
        private :
            WSDL::WebService wsdl;  // web service definition
            string portType;   // service name

            __7_ code log;
            __7_ code dbglog;

public:
        //! @endcond

        public :
            //! target URL
            string url;
            //! HTTP headers to use
            hash headers = Headers;

public:

        //! creates the object based on a %WSDL which is parsed to a @ref WSDL::WebService "WebService" object which provides the basis for all communication with this object
/** one of either the \c wsdl or \c wsdl_file keys is required in the hash given to the constructor or an exception will be thrown
            @param h valid option keys:
            - \c wsdl: the URL of the web service or a @ref WSDL::WebService "WebService" object itself
            - \c wsdl_file: a path to use to load the %WSDL and create the @ref WSDL::WebService "WebService" object
            - \c url: override the target URL given in the %WSDL
            - [\c portType]: in case multiple portType entries are found in the WSDL, give the one to be used here
            - [\c port]: in case multiple port entries are found in the WSDL, give the one to be used here
            - [\c log]: a log closure or call reference taking a single string giving the log message
            - [\c dbglog]: a log closure or call reference taking a single string giving the debug log message
            - also all options from SoapClient::SoapClient::HTTPOptions, which are passed to the HTTPClient constructor
        */
        constructor(hash h);


        //! returns a hash representing the serialized SOAP request for a given @ref WSDL::WSOperation "WSOperation"
/** the returned hash can be passed to makeXMLString() to make the actual SOAP message

            @param operation the SOAP operation to use to serialize the request; if the operation is not known to the underlying @ref WSDL::WebService "WebService" class, an exception will be thrown
            @param args the arguments to the SOAP operation
            @param header data structure for the SOAP header, if required by the message
            @param op a reference to return the @ref WSDL::WSOperation "WSOperation" object found

            @throw SOAP-CLIENT-ERROR the operation is not defined in the WSDL
        */
        hash getMsg(string operation, any args, __7_ hash header, reference op, __7_ hash nsh);


        //! makes a server call with the given operation and arguments and returns the deserialized result
/** @param operation the operation name for the SOAP call
            @param args the arguments to the SOAP operation
            @param header optional soap headers (if required by the operation)

            @return the deserialized result of the SOAP call to the SOAP server
        */
        any call(string operation, any args, __7_ hash header, __7_ hash nsh);


        //! makes a server call with the given operation and arguments and returns the deserialized result
/** @param operation the operation name for the SOAP call
            @param args the arguments to the SOAP operation
            @param info an optional reference to return a hash of technical information about the SOAP call (raw message info and headers)

            @return the deserialized result of the SOAP call to the SOAP server
        */
        any call(string operation, any args, __7_ reference info);


        //! makes a server call with the given operation and arguments and returns the deserialized result with an output argument giving technical information about the call
/** @param info an optional reference to return a hash of technical information about the SOAP call (raw message info and headers)
            @param operation the operation name for the SOAP call
            @param args the arguments to the SOAP operation
            @param header optional soap headers (if required by the operation)

            @return the deserialized result of the SOAP call to the SOAP server
        */
        any call(reference info, string operation, any args, __7_ hash header, __7_ hash nsh);


        //! makes the call to the SOAP server and ensures that SOAP fault responses returned with a 500-series status code are processed as a SOAP fault so that error information is returned in the resulting exception
        private any makeCallIntern(__7_ reference info, string operation, any args, __7_ hash header, __7_ hash nsh);


        //! uses SoapClient::call() to transparently serialize the argument and make a call to the given operation and return the deserialized results
/** @param op the operation name, which is the method name passed to methodGate()
            @param arg a list or arguments or a single argument (or NOTHING) for the operation

            @return the deserialized result of the SOAP call to the SOAP server
        */
        any methodGate(string op, any arg);


        //! returns a hash that can be used to ensure serialization with the XSD type given as the \a type argument
        hash getType(string type, any v);

       
        //! sends a log message to the log closure or call reference, if any
        log(string msg);


        //! sends a log message to the debug log closure or call reference, if any
        dbglog(string msg);

    };
};
